import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;

/**
 * Run frequent gap filter and create report of filtered-out contacts.
 * @author Pawel P. Wozniak
 */
public class FrequentGaps {
	/**
	 * @param args - 
	 * [0]: path to FASTA file; 
	 * [1]: path to gplmDCA results; 
	 * [2]: path to MSA file; 
	 * [3]: value of Tgaps parameter (double; default = 95); 
	 * [4]: number of predicted contacts (N; default = 200); 
	 * [5]: for number of predicted contacts value of 'm' in 'm*L/d' where 'L' is sequence length (if 'm' is set to non-0 then [4] is ignored); 
	 * [6]: for number of predicted contacts value of 'd' in 'm*L/d' where 'L' is sequence length (default = 1); 
	 * [7]: chosen sub-name for the output file;
	 */
	public static void main(String[] args) {
		try
		{
			String fastaPath = ""; //path to the FASTA file (first from the file will be read)
			String aliPath = ""; //path to the file with alignment
			String resPath = ""; //path to the file with DCA results
			String upperTitle = ""; //subtitle for the file with filter output
			
			//default filters parameters (below): 
			double gapFreq = 95; //value of Tgaps parameter for frequent gap filter
			boolean ifBothGapsFreq = false; //True if both residues in pair should be in agreement with Tgaps parameter
						
			int top = 200; //number of predicted contacts N to apply filtration on (if mult=0 and div=0)
			double mult = 0; //if number of predicted contacts to apply filtration on depends on sequence length then e.g. 3L/2 == mult*L/div
			double div = 1; //if number of predicted contacts to apply filtration on depends on sequence length then e.g. 3L/2 == mult*L/div
			
			int separation = 5; //separation parameter for contact definition
			//default filters parameters (above): 
			
			//set input parameters (below):
			try
			{
				if(args.length < 3)
				{
					System.out.println("You must give at least 3 parameters - paths to FASTA, DCA results, and MSA.");
					return;
				}
				else
				{
					for (int i = 0; i < args.length; i++) {
						if(i == 0)
						{
							fastaPath = args[0];
							System.out.println("FASTA path: " + fastaPath);
						}
						else if(i == 1)
						{
							resPath = args[1];
							System.out.println("DCA results path: " + resPath);
						}
						else if(i == 2)
						{
							aliPath = args[2];
							System.out.println("MSA path: " + aliPath);
						}
						else if(i == 3)
						{
							gapFreq = Double.parseDouble(args[3]);
							System.out.println("Tgaps parameter: " + gapFreq);
						}
						else if(i == 4)
						{
							top = Integer.parseInt(args[4]);
							System.out.println("N: " + top);
						}
						else if(i == 5)
						{
							mult = Double.parseDouble(args[5]);
							if(mult != 0)
							{
								System.out.println("'m' in 'm*L/d': " + mult + " (N value ignored)");
							}
						}
						else if(i == 6)
						{
							div = Double.parseDouble(args[6]);
							if(mult != 0)
							{
								System.out.println("'d' in 'm*L/d': " + div);
							}
						}
						else if(i == 7)
						{
							upperTitle = args[7];
							System.out.println("Output sub-name: " + upperTitle);
						}
					}
				}
			}
			catch(Exception ex)
			{
				System.out.println("There was a problem with input values (probably the last not printed-out)" +
						" - see documentation and error report below.");
				ex.printStackTrace();
				return;
			}
			//set input parameters (above):
			
			//creating title with number of N and name of selected filter (below):
			String titleSuff = upperTitle;
			if(mult > 0)
			{
				titleSuff = titleSuff + "_topM" + ((int)mult) + "D" + ((int)div);
			}
			else
			{
				titleSuff = titleSuff + "_top" + top;
			}
			titleSuff = titleSuff + "_frequent-gap_Tgaps" + gapFreq;
			//creating title with number of N and name of selected filter (above):

			//initialize writer and list for filtered-out contacts:
			FileWriter writerReport = new FileWriter("raport_" + titleSuff);
			List<String> raport = new ArrayList<String>();
			
			//get FASTA sequence (first from the file):
			String fastaSeq = new FastaFileReader(fastaPath, 1).getFastaSeqs().get(0);
			
			//get number of of predicted contacts N to apply filtration on:
			int topTmp = top;
			if(mult > 0)
			{
				topTmp = (int) ((mult*fastaSeq.length())/div);
			}
			
			//run filtering:
			createReport(resPath, topTmp, separation,
					fastaSeq, raport, aliPath,
					ifBothGapsFreq, gapFreq);
			
			//create file with filtered-out contacts:
			for (int j = 0; j < raport.size(); j++) {
				writerReport.write(raport.get(j) + "\n");
			}
			writerReport.close();
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}

	/**
	 * Run filtering and create report of filtered-out contacts.
	 * @param resPath - path to the file with DCA results
	 * @param top - number of of predicted contacts N to apply filtration on
	 * @param separation - separation parameter for contact definition
	 * @param fastaSeq - FASTA sequence
	 * @param raport - variable which will contain the list of filtered-out contacts
	 * @param aliPath - path to the file with alignment
	 * @param ifBothGapsFreq - True if both residues in pair should be in agreement with Tgaps parameter
	 * @param gapFreq - value of Tgaps parameter for frequent gap filter
	 */	
	private static void createReport(String resPath, int top, int separation,
			String fastaSeq, List<String> raport, String aliPath,
			boolean ifBothGapsFreq, double gapFreq) {
		try
		{
			//Get sorted results from DCA file:
			TopMethods topMeth = new TopMethods();
			topMeth.gatherSortedGplmdca(resPath, 0, 1, 2, ",");
			int[] gdcaSortInd = topMeth.getGdcaSortInd();
			double[] gdcaSortVal = topMeth.getGdcaSortVal();
			List<Integer> posDca1 = topMeth.getPosDca1();
			List<Integer> posDca2 = topMeth.getPosDca2();

			//counters for number of residue pairs to apply filtering on:
			int counter = 0;
			int counterOriTop = 0;

			//reading MSA if needed (for entropy, gaps, and conservation filters):
			FastaFileReader fastAliReader = new FastaFileReader(aliPath, 1000000);
			MSAanalysis msaAnalysis = new MSAanalysis(fastAliReader.getFastaSeqs());
			
			//reading strongest N predicted contacts:
			double maxDca = -1000;
			double minDca = 1000;
			for(int i=0; i<gdcaSortInd.length; i++)
			{					
				int pos1 = posDca1.get(gdcaSortInd[i]);
				int pos2 = posDca2.get(gdcaSortInd[i]);
				double dcaScore = gdcaSortVal[i];
				
				boolean pass = true;				
				if(Math.abs(pos1-pos2) >= separation)
				{
					if(counterOriTop == 0 && maxDca == -1000)
					{
						maxDca = dcaScore;
					}
					
					if(counterOriTop == top && minDca == 1000)
					{
						minDca = dcaScore;
					}
					
					pass = ifGapsFreqPass(msaAnalysis, pos1, pos2, ifBothGapsFreq, gapFreq);
					
					counterOriTop++;
					if(pass)
					{
						counter++;
					}
					
					if(!pass)
					{
						raport.add(pos1 + " " + pos2 + " " + dcaScore + "     " + maxDca + " " + counterOriTop);
					}
					
					if(counter >= top)
					{
						break;
					}
				}
			}
			raport.add("MinDCA: " + minDca);
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}

	private static boolean ifGapsFreqPass(MSAanalysis msaAnalysis, int pos1, int pos2, boolean ifBothGapsFreq, double gapFreq) {
		boolean gap1 = !msaAnalysis.checkIfNonDeletion(pos1-1, gapFreq);
		boolean gap2 = !msaAnalysis.checkIfNonDeletion(pos2-1, gapFreq);
		
		if((ifBothGapsFreq && gap1 && gap2) || (!ifBothGapsFreq && (gap1 || gap2)))
		{
			return false;
		}

		return true;
	}
}
