import java.util.List;

/**
 * Calculates some MSA characteristics for positions, e.g. entropy, variability, deletion number.
 * @author Pawel P. Wozniak
 */
public class MSAanalysis
{
	private int[][] Alignment = null;
	private int AliNmb = 0;
	private int aaSymbols = 21;
	
	/**
	 * Get alignment in the integer format.
	 * @return alignment in the integer format where each amino acid is a different number
	 */
	public int[][] getAlignment()
	{
		return Alignment.clone();
	}
	
	/**
	 * @param Align - matrix [sequence number x sequence length] of numbers representing different amino acids
	 */
	MSAanalysis(int[][] Align)
	{
		Alignment = Align.clone();
		AliNmb = Align.length;
	}
	
	/**
	 * @param Align - list of sequences written as strings of amino acids symbols
	 */
	MSAanalysis(List<String> Align) 
	{
		AliNmb = Align.size();
		int SeqLen = Align.get(0).length();
		Alignment = new int[AliNmb][SeqLen];
		for(int i=0; i<AliNmb; i++)
		{
			for(int j=0; j<SeqLen; j++)
			{				
				switch (Align.get(i).charAt(j)) 
				{
					case '-':
						Alignment[i][j] = 0;
						break;
					case 'A':
						Alignment[i][j] = 1;
						break;
					case 'C':
						Alignment[i][j] = 2;
						break;
					case 'D':
						Alignment[i][j] = 3;
						break;
					case 'E':
						Alignment[i][j] = 4;
						break;
					case 'F':
						Alignment[i][j] = 5;
						break;
					case 'G':
						Alignment[i][j] = 6;
						break;
					case 'H':
						Alignment[i][j] = 7;
						break;
					case 'I':
						Alignment[i][j] = 8;
						break;
					case 'K':
						Alignment[i][j] = 9;
						break;
					case 'L':
						Alignment[i][j] = 10;
						break;
					case 'M':
						Alignment[i][j] = 11;
						break;
					case 'N':
						Alignment[i][j] = 12;
						break;
					case 'P':
						Alignment[i][j] = 13;
						break;
					case 'Q':
						Alignment[i][j] = 14;
						break;
					case 'R':
						Alignment[i][j] = 15;
						break;
					case 'S':
						Alignment[i][j] = 16;
						break;
					case 'T':
						Alignment[i][j] = 17;
						break;
					case 'V':
						Alignment[i][j] = 18;
						break;
					case 'W':
						Alignment[i][j] = 19;
						break;
					case 'Y':
						Alignment[i][j] = 20;
						break;
					default:
						Alignment[i][j] = 0; //Everything not included in 21 symbols will be counted as a gap
						break;
				} 
			}
		}
	}
	
	/**
	 * Check if there is deletion or unknown amino acid frequency higher than the cutoff at the chosen position in MSA.
	 * @param i - position in the MSA (numbered from 0)
	 * @param cutoff - cutoff frequency in percentage
	 * @return true if deletion or unknown amino acid frequency is lower (or equal) than the cutoff at the chosen position
	 */
	public boolean checkIfNonDeletion(int i, double cutoff)
	{
		int counter = 0;
		for(int p=0; p<AliNmb; p++) //Iteration for the first alignment
		{
			if(Alignment[p][i] == 0)
			{
				counter++;
			}
		}
		
		if((((double)counter/(double)AliNmb)*100.0) > cutoff)
		{
			return false;
		}
		else
		{
			return true;
		}
	}
	
	/**
	 * Calculate Shannon Entropy.
	 * @param i - position in the MSA (numbered from 0)
	 * @param ifGaps - true/false - with/without counting gaps
	 * @return Shannon Entropy value
	 */
	public double shannonEntropyCalc(int i, boolean ifGaps)
	{
		return runEntropyCalc(i, ifGaps, false);
	}
	
	/**
	 * Calculate Shannon Entropy.
	 * @param i - position in the MSA (numbered from 0)
	 * @param ifGaps - true/false - with/without counting gaps
	 * @param ifNormalized - true if result must be normalized
	 * @return Shannon Entropy value
	 */
	public double shannonEntropyCalc(int i, boolean ifGaps, boolean ifNormalized)
	{
		return runEntropyCalc(i, ifGaps, ifNormalized);
	}
	
	/**
	 * Calculate Shannon Entropy.
	 * @param i - position in the MSA (numbered from 0)
	 * @param ifGaps - true/false - with/without counting gaps
	 * @param ifNormalized - true if result must be normalized
	 * @return Shannon Entropy value
	 */
	private double runEntropyCalc(int i, boolean ifGaps, boolean ifNormalized)
	{
		long[] Buckets = new long[aaSymbols+1];
		for(int p=0; p<AliNmb; p++) //Iteration for the first alignment
		{
			if(ifGaps || Alignment[p][i] != 0)
			{
				Buckets[Alignment[p][i]]++;
				Buckets[aaSymbols]++;
			}
		}
		
		double Sp = 0;
		double fp = 0;
		int start;
		
		if(ifGaps)
		{
			start = 0;
		}
		else
		{
			start = 1;
		}
		
		for(int b=start; b<aaSymbols; b++)
		{
			fp = (double)Buckets[b]/(double)Buckets[aaSymbols];
			if(fp != 0)
			{
				if(ifNormalized)
				{
					Sp = Sp - (fp*(Math.log(fp)/Math.log(aaSymbols-start)));
				}
				else
				{
					Sp = Sp - (fp*Math.log(fp));
				}
			}
		}
				
		return Sp;
	}
	
	/**
	 * Calculate variability for the MSA position - number of different amino acids which occur more frequent than the cutoff value.
	 * @param i - position in the MSA (numbered from 0)
	 * @param cutoff - cutoff frequency in percentage
	 * @param ifGaps - true/false - with/without counting gaps
	 * @return variability value
	 */
	public int variabilityCalc(int i, double cutoff, boolean ifGaps)
	{
		long[] Buckets = new long[aaSymbols+1];
		for(int p=0; p<AliNmb; p++) //Iteration for the first alignment
		{
			if(ifGaps || Alignment[p][i] != 0)
			{
				Buckets[Alignment[p][i]]++;
				Buckets[aaSymbols]++;
			}
		}
		
		double fp = 0;
		int var = 0;
		int start;
		if(ifGaps)
		{
			start = 0;
		}
		else
		{
			start = 1;
		}
		
		for(int b=start; b<aaSymbols; b++)
		{
			fp = ((double)Buckets[b]/(double)Buckets[aaSymbols])*100.0;
			if(fp >= cutoff)
			{
				var++;
			}
		}
		
		return var;
	}
}
