import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Run the non-strongest Cys-Cys filter and create report of filtered-out contacts.
 * @author Pawel P. Wozniak
 */
public class NonStrongestCysCys {
	/**
	 * @param args - 
	 * [0]: path to FASTA file; 
	 * [1]: path to gplmDCA results; 
	 * [2]: number of predicted contacts (N; default = 200); 
	 * [3]: for number of predicted contacts value of 'm' in 'm*L/d' where 'L' is sequence length (if 'm' is set to non-0 then [2] is ignored); 
	 * [4]: for number of predicted contacts value of 'd' in 'm*L/d' where 'L' is sequence length (default = 1); 
	 * [5]: chosen sub-name for the output file;
	 */
	public static void main(String[] args) {
		try
		{
			String fastaPath = ""; //path to the FASTA file (first from the file will be read)
			String resPath = ""; //path to the file with DCA results
			String upperTitle = ""; //subtitle for the file with filter output
			
			//default filters parameters (below): 
			int top = 200; //number of predicted contacts N to apply filtration on (if mult=0 and div=0)
			double mult = 0; //if number of predicted contacts to apply filtration on depends on sequence length then e.g. 3L/2 == mult*L/div
			double div = 1; //if number of predicted contacts to apply filtration on depends on sequence length then e.g. 3L/2 == mult*L/div
			
			int separation = 5; //separation parameter for contact definition
			//default filters parameters (above): 
			
			//set input parameters (below):
			try
			{
				if(args.length < 2)
				{
					System.out.println("You must give at least 2 parameters - paths to FASTA and DCA results.");
					return;
				}
				else
				{
					for (int i = 0; i < args.length; i++) {
						if(i == 0)
						{
							fastaPath = args[0];
							System.out.println("FASTA path: " + fastaPath);
						}
						else if(i == 1)
						{
							resPath = args[1];
							System.out.println("DCA results path: " + resPath);
						}
						else if(i == 2)
						{
							top = Integer.parseInt(args[2]);
							System.out.println("N: " + top);
						}
						else if(i == 3)
						{
							mult = Double.parseDouble(args[3]);
							if(mult != 0)
							{
								System.out.println("'m' in 'm*L/d': " + mult + " (N value ignored)");
							}
						}
						else if(i == 4)
						{
							div = Double.parseDouble(args[4]);
							if(mult != 0)
							{
								System.out.println("'d' in 'm*L/d': " + div);
							}
						}
						else if(i == 5)
						{
							upperTitle = args[5];
							System.out.println("Output sub-name: " + upperTitle);
						}
					}
				}
			}
			catch(Exception ex)
			{
				System.out.println("There was a problem with input values (probably the last not printed-out)" +
						" - see documentation and error report below.");
				ex.printStackTrace();
				return;
			}
			//set input parameters (above):
			
			//creating title with number of N and name of selected filter (below):
			String titleSuff = upperTitle;
			if(mult > 0)
			{
				titleSuff = titleSuff + "_topM" + ((int)mult) + "D" + ((int)div);
			}
			else
			{
				titleSuff = titleSuff + "_top" + top;
			}
			titleSuff = titleSuff + "_non-strongest-Cys-Cys";
			//creating title with number of N and name of selected filter (above):

			//initialize writer and list for filtered-out contacts:
			FileWriter writerReport = new FileWriter("raport_" + titleSuff);
			List<String> raport = new ArrayList<String>();
			
			//get FASTA sequence (first from the file):
			String fastaSeq = new FastaFileReader(fastaPath, 1).getFastaSeqs().get(0);
			
			//get number of of predicted contacts N to apply filtration on:
			int topTmp = top;
			if(mult > 0)
			{
				topTmp = (int) ((mult*fastaSeq.length())/div);
			}
			
			//run filtering:
			createReport(resPath, topTmp, separation, fastaSeq, raport);
			
			//create file with filtered-out contacts:
			for (int j = 0; j < raport.size(); j++) {
				writerReport.write(raport.get(j) + "\n");
			}
			writerReport.close();
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}

	/**
	 * Run filtering and create report of filtered-out contacts.
	 * @param resPath - path to the file with DCA results
	 * @param top - number of of predicted contacts N to apply filtration on
	 * @param separation - separation parameter for contact definition
	 * @param fastaSeq - FASTA sequence
	 * @param raport - variable which will contain the list of filtered-out contacts
	 */	
	private static void createReport(String resPath, int top, int separation, String fastaSeq, List<String> raport) {
		try
		{
			//Get sorted results from DCA file:
			TopMethods topMeth = new TopMethods();
			topMeth.gatherSortedGplmdca(resPath, 0, 1, 2, ",");
			int[] gdcaSortInd = topMeth.getGdcaSortInd();
			double[] gdcaSortVal = topMeth.getGdcaSortVal();
			List<Integer> posDca1 = topMeth.getPosDca1();
			List<Integer> posDca2 = topMeth.getPosDca2();

			//counters for number of residue pairs to apply filtering on:
			int counter = 0;
			int counterOriTop = 0;
			
			//list of Cys-Cys pairs for non-strongest Cys-Cys filter
			List<String> alreadyFound = new ArrayList<String>();
			
			//reading strongest N predicted contacts:
			double maxDca = -1000;
			double minDca = 1000;
			for(int i=0; i<gdcaSortInd.length; i++)
			{					
				int pos1 = posDca1.get(gdcaSortInd[i]);
				int pos2 = posDca2.get(gdcaSortInd[i]);
				String aa1 = String.valueOf(fastaSeq.charAt(pos1-1));
				String aa2 = String.valueOf(fastaSeq.charAt(pos2-1));
				double dcaScore = gdcaSortVal[i];
				
				//run the non-strongest Cys-Cys filter:
				boolean pass = ifFirstCysMarksPass(pos1, pos2, aa1, aa2, alreadyFound);
				
				if(Math.abs(pos1-pos2) >= separation)
				{
					if(counterOriTop == 0 && maxDca == -1000)
					{
						maxDca = dcaScore;
					}
					
					if(counterOriTop == top && minDca == 1000)
					{
						minDca = dcaScore;
					}

					counterOriTop++;
					if(pass)
					{
						counter++;
					}
					
					if(!pass)
					{
						raport.add(pos1 + " " + pos2 + " " + dcaScore + "     " + maxDca + " " + counterOriTop);
					}
					
					if(counter >= top)
					{
						break;
					}
				}
			}
			raport.add("MinDCA: " + minDca);
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}

	private static boolean ifFirstCysMarksPass(int pos1, int pos2, String aa1, String aa2, List<String> alreadyFound) {
		boolean ifFirstFound = false;
		boolean ifSecondFound = false;
		if(aa1.equals("C") && aa2.equals("C"))
		{
			ifFirstFound = checkIfFound(alreadyFound, String.valueOf(pos1)+aa1);
			ifSecondFound = checkIfFound(alreadyFound, String.valueOf(pos2)+aa2);
			if(!ifFirstFound)
			{
				alreadyFound.add(String.valueOf(pos1)+aa1);
			}
			if(!ifSecondFound)
			{
				alreadyFound.add(String.valueOf(pos2)+aa2);
			}
			if(ifFirstFound || ifSecondFound)
			{
				return false;
			}
		}
		return true;
	}

	private static boolean checkIfFound(List<String> alreadyFound, String toFind) {
		for (Iterator<String> iterator = alreadyFound.iterator(); iterator.hasNext();) {
			String whatnow = iterator.next();
			if(whatnow.equals(toFind))
			{
				return true;
			}
		}
		return false;
	}
}
