import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;

/**
 * Run secondary structure straddling filter and create report of filtered-out contacts.
 * @author Pawel P. Wozniak
 */
public class SecondaryStraddling {
	/**
	 * @param args - 
	 * [0]: path to FASTA file; 
	 * [1]: path to gplmDCA results; 
	 * [2]: path to SSpro results file; 
	 * [3]: number of predicted contacts (N; default = 200); 
	 * [4]: for number of predicted contacts value of 'm' in 'm*L/d' where 'L' is sequence length (if 'm' is set to non-0 then [3] is ignored); 
	 * [5]: for number of predicted contacts value of 'd' in 'm*L/d' where 'L' is sequence length (default = 1); 
	 * [6]: chosen sub-name for the output file;
	 */
	public static void main(String[] args) {
		try
		{
			String ssPredPath = ""; //path to the file with SSpro results
			String fastaPath = ""; //path to the FASTA file (first from the file will be read)
			String resPath = ""; //path to the file with DCA results
			String upperTitle = ""; //subtitle for the file with filter output
			
			//default filters parameters (below):
			String notAllowed = "C"; //these secondary structure will not be taken under consideration when list of secondary structure fragments is created for a sequence
			int anyfragSize = 2; //only secondary structure fragments of minimally that length will be used when list of secondary structure fragments is created for a sequence
			
			int top = 200; //number of predicted contacts N to apply filtration on (if mult=0 and div=0)
			double mult = 0; //if number of predicted contacts to apply filtration on depends on sequence length then e.g. 3L/2 == mult*L/div
			double div = 1; //if number of predicted contacts to apply filtration on depends on sequence length then e.g. 3L/2 == mult*L/div
			
			int separation = 5; //separation parameter for contact definition
			//default filters parameters (above): 
			
			//set input parameters (below):
			try
			{
				if(args.length < 3)
				{
					System.out.println("You must give at least 3 parameters - paths to FASTA, DCA results, and SSpro results.");
					return;
				}
				else
				{
					for (int i = 0; i < args.length; i++) {
						if(i == 0)
						{
							fastaPath = args[0];
							System.out.println("FASTA path: " + fastaPath);
						}
						else if(i == 1)
						{
							resPath = args[1];
							System.out.println("DCA results path: " + resPath);
						}
						else if(i == 2)
						{
							ssPredPath = args[2];
							System.out.println("SSpro results path: " + ssPredPath);
						}
						else if(i == 3)
						{
							top = Integer.parseInt(args[3]);
							System.out.println("N: " + top);
						}
						else if(i == 4)
						{
							mult = Double.parseDouble(args[4]);
							if(mult != 0)
							{
								System.out.println("'m' in 'm*L/d': " + mult + " (N value ignored)");
							}
						}
						else if(i == 5)
						{
							div = Double.parseDouble(args[5]);
							if(mult != 0)
							{
								System.out.println("'d' in 'm*L/d': " + div);
							}
						}
						else if(i == 6)
						{
							upperTitle = args[6];
							System.out.println("Output sub-name: " + upperTitle);
						}
					}
				}
			}
			catch(Exception ex)
			{
				System.out.println("There was a problem with input values (probably the last not printed-out)" +
						" - see documentation and error report below.");
				ex.printStackTrace();
				return;
			}
			//set input parameters (above):
			
			//creating title with number of N and name of selected filter (below):
			String titleSuff = upperTitle;
			if(mult > 0)
			{
				titleSuff = titleSuff + "_topM" + ((int)mult) + "D" + ((int)div);
			}
			else
			{
				titleSuff = titleSuff + "_top" + top;
			}
			titleSuff = titleSuff + "_secondary-structure-straddling";
			//creating title with number of N and name of selected filter (above):

			//initialize writer and list for filtered-out contacts:
			FileWriter writerReport = new FileWriter("raport_" + titleSuff);
			List<String> raport = new ArrayList<String>();
			
			//get FASTA sequence (first from the file):
			String fastaSeq = new FastaFileReader(fastaPath, 1).getFastaSeqs().get(0);
			
			//get number of of predicted contacts N to apply filtration on:
			int topTmp = top;
			if(mult > 0)
			{
				topTmp = (int) ((mult*fastaSeq.length())/div);
			}
			
			//run filtering:
			createReport(resPath, topTmp, separation, anyfragSize, notAllowed,
					ssPredPath, fastaSeq, raport);
			
			//create file with filtered-out contacts:
			for (int j = 0; j < raport.size(); j++) {
				writerReport.write(raport.get(j) + "\n");
			}
			writerReport.close();
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}

	/**
	 * Run filtering and create report of filtered-out contacts.
	 * @param resPath - path to the file with DCA results
	 * @param top - number of of predicted contacts N to apply filtration on
	 * @param separation - separation parameter for contact definition
	 * @param anyfragSize - only secondary structure fragments of minimally that length will be used when list of secondary structure fragments is created for a sequence
	 * @param notAllowed - these secondary structure will not be taken under consideration when list of secondary structure fragments is created for a sequence
	 * @param ssPredPath - path to the file with SSpro results
	 * @param fastaSeq - FASTA sequence
	 * @param raport - variable which will contain the list of filtered-out contacts
	 */	
	private static void createReport(String resPath, int top, int separation, int anyfragSize, String notAllowed,
			String ssPredPath, String fastaSeq, List<String> raport) {
		try
		{
			//Get sorted results from DCA file:
			TopMethods topMeth = new TopMethods();
			topMeth.gatherSortedGplmdca(resPath, 0, 1, 2, ",");
			int[] gdcaSortInd = topMeth.getGdcaSortInd();
			double[] gdcaSortVal = topMeth.getGdcaSortVal();
			List<Integer> posDca1 = topMeth.getPosDca1();
			List<Integer> posDca2 = topMeth.getPosDca2();

			//counters for number of residue pairs to apply filtering on:
			int counter = 0;
			int counterOriTop = 0;
			
			//list of secondary structure fragments:
			List<int[]> frags = new ArrayList<int[]>();
			List<String> fragSS = new ArrayList<String>();
			
			//getting data from SSpro and ACCpro20:
			List<Character> ssPred = null;
			File ssPredFile = new File(ssPredPath);
			if(ssPredFile.exists())
			{
				ssPred = ScratchReader.ssRead(ssPredFile.getAbsolutePath(), null);
			}
			else
			{
				System.out.println("No " + ssPredPath + " !!!");
			}
			
			if(ssPred != null)
			{
				//get list of secondary structure fragments:
				getFragsForPredicted(ssPredPath, fastaSeq, anyfragSize, frags, fragSS, notAllowed, ssPred);
				
				//reading strongest N predicted contacts:
				double maxDca = -1000;
				double minDca = 1000;
				for(int i=0; i<gdcaSortInd.length; i++)
				{					
					int pos1 = posDca1.get(gdcaSortInd[i]);
					int pos2 = posDca2.get(gdcaSortInd[i]);
					double dcaScore = gdcaSortVal[i];
					
					boolean pass = true;					
					if(Math.abs(pos1-pos2) >= separation)
					{
						if(counterOriTop == 0 && maxDca == -1000)
						{
							maxDca = dcaScore;
						}
						
						if(counterOriTop == top && minDca == 1000)
						{
							minDca = dcaScore;
						}
						
						pass = ifPassIntraSSMarks(frags, fragSS, pos1, pos2, separation);
						
						counterOriTop++;
						if(pass)
						{
							counter++;
						}
						
						if(!pass)
						{
							raport.add(pos1 + " " + pos2 + " " + dcaScore + "     " + maxDca + " " + counterOriTop);
						}
						
						if(counter >= top)
						{
							break;
						}
					}
				}
				raport.add("MinDCA: " + minDca);
			}
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}

	private static boolean ifPassIntraSSMarks(List<int[]> frags, List<String> fragSS, int pos1, int pos2, int separation) {
		if(Math.abs(pos1-pos2) >= separation)
		{
			for (int i = 0; i < frags.size(); i++) {
				if(fragSS.get(i).contains("H"))
				{
					if((pos1 >= frags.get(i)[0]-3 && pos1 <= frags.get(i)[1] && pos2 >= frags.get(i)[0]-3 && pos2 <= frags.get(i)[1]) ||
						((pos1 == frags.get(i)[0]-1 && pos2 == frags.get(i)[1]+1) || (pos2 == frags.get(i)[0]-1 && pos1 == frags.get(i)[1]+1)))
					{
						return false;
					}
				}
				else if(fragSS.get(i).contains("E"))
				{
					if(
						(pos1 >= frags.get(i)[0]-4 && pos1 <= frags.get(i)[1] && pos2 >= frags.get(i)[0]-4 && pos2 <= frags.get(i)[1]) ||
						((pos1 == frags.get(i)[0]-1 && pos2 == frags.get(i)[1]+1) || (pos2 == frags.get(i)[0]-1 && pos1 == frags.get(i)[1]+1)) ||
						(
							Math.abs(frags.get(i)[0]-frags.get(i)[1]) >= 5 && 
							(
								(pos1 == frags.get(i)[0]-2 && pos2 == frags.get(i)[1]+2) || (pos2 == frags.get(i)[0]-2 && pos1 == frags.get(i)[1]+2)
							)
						)
					  )
					{
						return false;
					}
				}
			}
		}
		return true;
	}

	private static void getFragsForPredicted(String ssPredPath, String fastaSeq, int fragSize, List<int[]> frags, List<String> fragSS, String notAllowed, List<Character> ssPred)
	{
		File ssPredFile = new File(ssPredPath);
		if(fastaSeq.length() == ssPred.size())
		{
			int counter = 0;
			int firstNmb = 0;
			char prevSS = '@';
			for (int i = 0; i < ssPred.size(); i++) {
				if(prevSS != ssPred.get(i))
				{
					if(notAllowed.isEmpty() || (!notAllowed.contains(String.valueOf(prevSS))))
					{
						if(counter >= fragSize)
						{
							int[] tmpFrag = new int[2];
							tmpFrag[0] = firstNmb;
							tmpFrag[1] = i;
							frags.add(tmpFrag);
							fragSS.add(String.valueOf(prevSS));
						}
					}
					prevSS = ssPred.get(i);
					counter = 1;
					firstNmb = i+1;
				}
				else
				{
					counter++;
				}
			}
			
			if(notAllowed.isEmpty() || (!notAllowed.contains(String.valueOf(prevSS))))
			{
				if(counter >= fragSize)
				{
					int[] tmpFrag = new int[2];
					tmpFrag[0] = firstNmb;
					tmpFrag[1] = ssPred.size();
					frags.add(tmpFrag);
					fragSS.add(String.valueOf(prevSS));
				}
			}			
		}
		else
		{
			System.out.println("Wrong SS length: " + ssPredFile.getName().split("\\.")[0]);
		}
	}
}
