import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;

/**
 * Run terminus filter and create report of filtered-out contacts.
 * @author Pawel P. Wozniak
 */
public class Terminus {
	/**
	 * @param args - 
	 * [0]: path to FASTA file; 
	 * [1]: path to gplmDCA results; 
	 * [2]: path to ACCpro20 results; 
	 * [3]: value of t parameter (int; default = 15); 
	 * [4]: value of TRSA parameter (double; default = 35); 
	 * [5]: number of predicted contacts (N; default = 200); 
	 * [6]: for number of predicted contacts value of 'm' in 'm*L/d' where 'L' is sequence length (if 'm' is set to non-0 then [5] is ignored); 
	 * [7]: for number of predicted contacts value of 'd' in 'm*L/d' where 'L' is sequence length (default = 1); 
	 * [8]: chosen sub-name for the output file;
	 */
	public static void main(String[] args) {
		try
		{
			String accPredPath = ""; //path to the file with ACCpro20 results
			String fastaPath = ""; //path to the FASTA file (first from the file will be read)
			String resPath = ""; //path to the file with DCA results
			String upperTitle = ""; //subtitle for the file with filter output
			
			//default filters parameters (below): 
			int howMany = 15; //value of t parameter for terminus filter
			boolean ifACCpasser = true; //True if TRSA parameter should be used for terminus filter
			boolean ifBothACC = false; //True if both residues in pair should be in agreement with TRSA parameter
			double accThresLE = 35; //value of TRSA parameter for terminus filter
			
			int top = 200; //number of predicted contacts N to apply filtration on (if mult=0 and div=0)
			double mult = 0; //if number of predicted contacts to apply filtration on depends on sequence length then e.g. 3L/2 == mult*L/div
			double div = 1; //if number of predicted contacts to apply filtration on depends on sequence length then e.g. 3L/2 == mult*L/div
			
			int separation = 5; //separation parameter for contact definition
			//default filters parameters (above): 
			
			//set input parameters (below):
			try
			{
				if(args.length < 3)
				{
					System.out.println("You must give at least 3 parameters - paths to FASTA, DCA results, and ACCpro20.");
					return;
				}
				else
				{
					for (int i = 0; i < args.length; i++) {
						if(i == 0)
						{
							fastaPath = args[0];
							System.out.println("FASTA path: " + fastaPath);
						}
						else if(i == 1)
						{
							resPath = args[1];
							System.out.println("DCA results path: " + resPath);
						}
						else if(i == 2)
						{
							accPredPath = args[2];
							System.out.println("ACCpro20 results path: " + accPredPath);
						}
						else if(i == 3)
						{
							howMany = Integer.parseInt(args[3]);
							System.out.println("t parameter: " + howMany);
						}
						else if(i == 4)
						{
							accThresLE = Double.parseDouble(args[4]);
							System.out.println("TRSA parameter: " + accThresLE);
						}
						else if(i == 5)
						{
							top = Integer.parseInt(args[5]);
							System.out.println("N: " + top);
						}
						else if(i == 6)
						{
							mult = Double.parseDouble(args[6]);
							if(mult != 0)
							{
								System.out.println("'m' in 'm*L/d': " + mult + " (N value ignored)");
							}
						}
						else if(i == 7)
						{
							div = Double.parseDouble(args[7]);
							if(mult != 0)
							{
								System.out.println("'d' in 'm*L/d': " + div);
							}
						}
						else if(i == 8)
						{
							upperTitle = args[8];
							System.out.println("Output sub-name: " + upperTitle);
						}
					}
				}
			}
			catch(Exception ex)
			{
				System.out.println("There was a problem with input values (probably the last not printed-out)" +
						" - see documentation and error report below.");
				ex.printStackTrace();
				return;
			}
			//set input parameters (above):
			
			//creating title with number of N and name of selected filter (below):
			String titleSuff = upperTitle;
			if(mult > 0)
			{
				titleSuff = titleSuff + "_topM" + ((int)mult) + "D" + ((int)div);
			}
			else
			{
				titleSuff = titleSuff + "_top" + top;
			}
			titleSuff = titleSuff + "_terminus_t" + howMany + "_TRSA" + accThresLE;
			//creating title with number of N and name of selected filter (above):

			//initialize writer and list for filtered-out contacts:
			FileWriter writerReport = new FileWriter("raport_" + titleSuff);
			List<String> raport = new ArrayList<String>();
			
			//get FASTA sequence (first from the file):
			String fastaSeq = new FastaFileReader(fastaPath, 1).getFastaSeqs().get(0);
			
			//get number of of predicted contacts N to apply filtration on:
			int topTmp = top;
			if(mult > 0)
			{
				topTmp = (int) ((mult*fastaSeq.length())/div);
			}
			
			//run filtering:
			createReport(resPath, topTmp, separation, accPredPath,
					fastaSeq, raport, howMany, ifACCpasser, ifBothACC, accThresLE);
			
			//create file with filtered-out contacts:
			for (int j = 0; j < raport.size(); j++) {
				writerReport.write(raport.get(j) + "\n");
			}
			writerReport.close();
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}

	/**
	 * Run filtering and create report of filtered-out contacts.
	 * @param resPath - path to the file with DCA results
	 * @param top - number of of predicted contacts N to apply filtration on
	 * @param separation - separation parameter for contact definition
	 * @param accPredPath - path to the file with ACCpro20 results
	 * @param fastaSeq - FASTA sequence
	 * @param raport - variable which will contain the list of filtered-out contacts
	 * @param howMany - value of t parameter for terminus filter
	 * @param ifACCpasser - True if TRSA parameter should be used for terminus filter
	 * @param ifBothACC - True if both residues in pair should be in agreement with TRSA parameter
	 * @param accThres - value of TRSA parameter for terminus filter
	 */	
	private static void createReport(String resPath, int top, int separation,
			String accPredPath, String fastaSeq, List<String> raport, int howMany,
			boolean ifACCpasser, boolean ifBothACC, double accThres) {
		try
		{
			//Get sorted results from DCA file:
			TopMethods topMeth = new TopMethods();
			topMeth.gatherSortedGplmdca(resPath, 0, 1, 2, ",");
			int[] gdcaSortInd = topMeth.getGdcaSortInd();
			double[] gdcaSortVal = topMeth.getGdcaSortVal();
			List<Integer> posDca1 = topMeth.getPosDca1();
			List<Integer> posDca2 = topMeth.getPosDca2();

			//counters for number of residue pairs to apply filtering on:
			int counter = 0;
			int counterOriTop = 0;
			
			//getting data from SSpro and ACCpro20:
			List<Integer> accPred = null;
			File accPredFile = new File(accPredPath);
			if(accPredFile.exists())
			{
				accPred = ScratchReader.accRead(accPredFile.getAbsolutePath(), null);
			}
			else
			{
				System.out.println("No " + accPredPath + " !!!");
			}
			
			if(accPred != null)
			{
				//reading strongest N predicted contacts:
				double maxDca = -1000;
				double minDca = 1000;
				for(int i=0; i<gdcaSortInd.length; i++)
				{					
					int pos1 = posDca1.get(gdcaSortInd[i]);
					int pos2 = posDca2.get(gdcaSortInd[i]);
					double dcaScore = gdcaSortVal[i];
					
					boolean pass = true;					
					if(Math.abs(pos1-pos2) >= separation)
					{
						if(counterOriTop == 0 && maxDca == -1000)
						{
							maxDca = dcaScore;
						}
						
						if(counterOriTop == top && minDca == 1000)
						{
							minDca = dcaScore;
						}
						
						boolean pass1 = getNmbOfContacts(pos1, pos2, 1, howMany, 1, fastaSeq.length());
						boolean pass2 = getNmbOfContacts(pos1, pos2, fastaSeq.length()-howMany+1, fastaSeq.length(), 1, fastaSeq.length());
						
						if(!pass1 || !pass2)
						{
							pass = false;
						}
						
						//apply TRSA for terminus filter:
						if(ifACCpasser && !pass)
						{
							pass = ifACCPass(pos1, pos2, accPred, accThres, ifBothACC);
						}
						
						counterOriTop++;
						if(pass)
						{
							counter++;
						}
						
						if(!pass)
						{
							raport.add(pos1 + " " + pos2 + " " + dcaScore + "     " + maxDca + " " + counterOriTop);
						}
						
						if(counter >= top)
						{
							break;
						}
					}
				}
				raport.add("MinDCA: " + minDca);
			}
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}

	private static boolean ifACCPass(int pos1, int pos2, List<Integer> accPred, double accThres, boolean ifBoth) {
		if((ifBoth && accPred.get(pos1-1) <= accThres && accPred.get(pos2-1) <= accThres) || (!ifBoth && (accPred.get(pos1-1) <= accThres || accPred.get(pos2-1) <= accThres)))
		{
			return true;
		}
		return false;
	}

	private static boolean getNmbOfContacts(int pos1, int pos2, int s1, int e1, int s2, int e2) {
		if((pos1 >= s1 && pos1 <= e1 && pos2 >= s2 && pos2 <= e2) || (pos2 >= s1 && pos2 <= e1 && pos1 >= s2 && pos1 <= e2))
		{
			return false;
		}
		return true;
	}
}
